import json
from typing import Any, Dict, List, Optional

from pydantic import TypeAdapter

from .._config import Config
from .._execution_context import ExecutionContext
from .._folder_context import FolderContext
from .._utils import Endpoint, RequestSpec, header_folder
from .._utils.constants import (
    HEADER_FOLDER_KEY,
    HEADER_FOLDER_PATH,
    ORCHESTRATOR_STORAGE_BUCKET_DATA_SOURCE,
)
from ..models import IngestionInProgressException
from ..models.context_grounding import ContextGroundingQueryResponse
from ..models.context_grounding_index import ContextGroundingIndex
from ..tracing._traced import traced
from ._base_service import BaseService
from .folder_service import FolderService


class ContextGroundingService(FolderContext, BaseService):
    """Service for managing semantic automation contexts in UiPath.

    Context Grounding is a feature that helps in understanding and managing the
    semantic context in which automation processes operate. It provides capabilities
    for indexing, retrieving, and searching through contextual information that
    can be used to enhance AI-enabled automation.

    This service requires a valid folder key to be set in the environment, as
    context grounding operations are always performed within a specific folder
    context.
    """

    def __init__(
        self,
        config: Config,
        execution_context: ExecutionContext,
        folders_service: FolderService,
    ) -> None:
        self._folders_service = folders_service
        super().__init__(config=config, execution_context=execution_context)

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def retrieve(
        self,
        name: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Optional[ContextGroundingIndex]:
        """Retrieve context grounding index information by its name.

        This method fetches details about a specific context index, which can be
        used to understand what type of contextual information is available for
        automation processes.

        Args:
            name (str): The name of the context index to retrieve.

        Returns:
            Optional[ContextGroundingIndex]: The index information, including its configuration and metadata if found, otherwise None.
        """
        spec = self._retrieve_spec(
            name,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = self.request(
            spec.method,
            spec.endpoint,
            params=spec.params,
            headers=spec.headers,
        ).json()
        return next(
            (
                ContextGroundingIndex.model_validate(item)
                for item in response["value"]
                if item["name"] == name
            ),
            None,
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def retrieve_async(
        self,
        name: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Optional[ContextGroundingIndex]:
        """Retrieve asynchronously context grounding index information by its name.

        This method fetches details about a specific context index, which can be
        used to understand what type of contextual information is available for
        automation processes.

        Args:
            name (str): The name of the context index to retrieve.

        Returns:
            Optional[ContextGroundingIndex]: The index information, including its configuration and metadata if found, otherwise None.

        """
        spec = self._retrieve_spec(
            name,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = (
            await self.request_async(
                spec.method,
                spec.endpoint,
                params=spec.params,
            )
        ).json()
        return next(
            (
                ContextGroundingIndex.model_validate(item)
                for item in response["value"]
                if item["name"] == name
            ),
            None,
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def retrieve_by_id(
        self,
        id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Any:
        """Retrieve context grounding index information by its ID.

        This method provides direct access to a context index using its unique
        identifier, which can be more efficient than searching by name.

        Args:
            id (str): The unique identifier of the context index.

        Returns:
            Any: The index information, including its configuration and metadata.
        """
        spec = self._retrieve_by_id_spec(
            id,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        return self.request(
            spec.method,
            spec.endpoint,
            params=spec.params,
        ).json()

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def retrieve_by_id_async(
        self,
        id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> Any:
        """Retrieve asynchronously context grounding index information by its ID.

        This method provides direct access to a context index using its unique
        identifier, which can be more efficient than searching by name.

        Args:
            id (str): The unique identifier of the context index.

        Returns:
            Any: The index information, including its configuration and metadata.

        """
        spec = self._retrieve_by_id_spec(
            id,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = await self.request_async(
            spec.method,
            spec.endpoint,
            params=spec.params,
        )

        return response.json()

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def search(
        self,
        name: str,
        query: str,
        number_of_results: int = 10,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[ContextGroundingQueryResponse]:
        """Search for contextual information within a specific index.

        This method performs a semantic search against the specified context index,
        helping to find relevant information that can be used in automation processes.
        The search is powered by AI and understands natural language queries.

        Args:
            name (str): The name of the context index to search in.
            query (str): The search query in natural language.
            number_of_results (int, optional): Maximum number of results to return.
                Defaults to 10.

        Returns:
            List[ContextGroundingQueryResponse]: A list of search results, each containing
                relevant contextual information and metadata.
        """
        index = self.retrieve(name)
        if index and index.in_progress_ingestion():
            raise IngestionInProgressException(index_name=name)

        spec = self._search_spec(
            name,
            query,
            number_of_results,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = self.request(
            spec.method,
            spec.endpoint,
            content=spec.content,
        )

        return TypeAdapter(List[ContextGroundingQueryResponse]).validate_python(
            response.json()
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def search_async(
        self,
        name: str,
        query: str,
        number_of_results: int = 10,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> List[ContextGroundingQueryResponse]:
        """Search asynchronously for contextual information within a specific index.

        This method performs a semantic search against the specified context index,
        helping to find relevant information that can be used in automation processes.
        The search is powered by AI and understands natural language queries.

        Args:
            name (str): The name of the context index to search in.
            query (str): The search query in natural language.
            number_of_results (int, optional): Maximum number of results to return.
                Defaults to 10.

        Returns:
            List[ContextGroundingQueryResponse]: A list of search results, each containing
                relevant contextual information and metadata.
        """
        index = self.retrieve(
            name,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        if index and index.in_progress_ingestion():
            raise IngestionInProgressException(index_name=name)
        spec = self._search_spec(
            name,
            query,
            number_of_results,
            folder_key=folder_key,
            folder_path=folder_path,
        )

        response = await self.request_async(
            spec.method,
            spec.endpoint,
            content=spec.content,
        )

        return TypeAdapter(List[ContextGroundingQueryResponse]).validate_python(
            response.json()
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def get_or_create_index(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        storage_bucket_name: str,
        file_name_glob: Optional[str] = None,
        storage_bucket_folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> ContextGroundingIndex:
        spec = self._create_spec(
            name,
            description,
            storage_bucket_name,
            file_name_glob,
            storage_bucket_folder_path,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        index = self.retrieve(
            name=name,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        if index:
            return index

        response = self.request(
            spec.method,
            spec.endpoint,
            content=spec.content,
            headers=spec.headers,
        ).json()
        return ContextGroundingIndex.model_validate(response)

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def get_or_create_index_async(
        self,
        name: str,
        *,
        description: Optional[str] = None,
        storage_bucket_name: str,
        file_name_glob: Optional[str] = None,
        storage_bucket_folder_path: Optional[str] = None,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> ContextGroundingIndex:
        index = await self.retrieve_async(
            name=name,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        if index:
            return index

        spec = self._create_spec(
            name,
            description,
            storage_bucket_name,
            file_name_glob,
            storage_bucket_folder_path,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        response = (
            await self.request_async(
                spec.method,
                spec.endpoint,
                content=spec.content,
                headers=spec.headers,
            )
        ).json()
        return ContextGroundingIndex.model_validate(response)

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def ingest_data(
        self,
        index: ContextGroundingIndex,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> None:
        if not index.id:
            return
        spec = self._ingest_spec(
            index.id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        self.request(
            spec.method,
            spec.endpoint,
            headers=spec.headers,
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def ingest_data_async(
        self,
        index: ContextGroundingIndex,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> None:
        if not index.id:
            return
        spec = self._ingest_spec(
            index.id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        await self.request_async(
            spec.method,
            spec.endpoint,
            headers=spec.headers,
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    def delete_index(
        self,
        index: ContextGroundingIndex,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> None:
        if not index.id:
            return
        spec = self._delete_by_id_spec(
            index.id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        self.request(
            spec.method,
            spec.endpoint,
            headers=spec.headers,
        )

    @traced(run_type="uipath", hide_input=True, hide_output=True)
    async def delete_index_async(
        self,
        index: ContextGroundingIndex,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> None:
        if not index.id:
            return
        spec = self._delete_by_id_spec(
            index.id,
            folder_key=folder_key,
            folder_path=folder_path,
        )
        await self.request_async(
            spec.method,
            spec.endpoint,
            headers=spec.headers,
        )

    @property
    def custom_headers(self) -> Dict[str, str]:
        self._folder_key = self._folder_key or (
            self._folders_service.retrieve_key_by_folder_path(self._folder_path)
            if self._folder_path
            else None
        )

        if self._folder_key is None:
            raise ValueError(
                f"Neither the folder key nor the folder path is set ({HEADER_FOLDER_KEY}, {HEADER_FOLDER_PATH})"
            )

        return self.folder_headers

    def _ingest_spec(
        self,
        key: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None

        return RequestSpec(
            method="POST",
            endpoint=Endpoint(f"/ecs_/v2/indexes/{key}/ingest"),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _retrieve_spec(
        self,
        name: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        print(folder_key, folder_path)
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None
        print("~~~", name, folder_key, folder_path)
        return RequestSpec(
            method="GET",
            endpoint=Endpoint("/ecs_/v2/indexes"),
            params={"$filter": f"Name eq '{name}'"},
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _create_spec(
        self,
        name: str,
        description: Optional[str],
        storage_bucket_name: Optional[str],
        file_name_glob: Optional[str],
        storage_bucket_folder_path: Optional[str],
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None

        storage_bucket_folder_path = (
            storage_bucket_folder_path
            if storage_bucket_folder_path
            else self._folder_path
        )
        return RequestSpec(
            method="POST",
            endpoint=Endpoint("/ecs_/v2/indexes/create"),
            content=json.dumps(
                {
                    "name": name,
                    "description": description,
                    "dataSource": {
                        "@odata.type": ORCHESTRATOR_STORAGE_BUCKET_DATA_SOURCE,
                        "folder": storage_bucket_folder_path,
                        "bucketName": storage_bucket_name,
                        "fileNameGlob": file_name_glob
                        if file_name_glob is not None
                        else "*",
                        "directoryPath": "/",
                    },
                }
            ),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _retrieve_by_id_spec(
        self,
        id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None

        return RequestSpec(
            method="GET",
            endpoint=Endpoint(f"/ecs_/v2/indexes/{id}"),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _delete_by_id_spec(
        self,
        id: str,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None

        return RequestSpec(
            method="DELETE",
            endpoint=Endpoint(f"/ecs_/v2/indexes/{id}"),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )

    def _search_spec(
        self,
        name: str,
        query: str,
        number_of_results: int = 10,
        folder_key: Optional[str] = None,
        folder_path: Optional[str] = None,
    ) -> RequestSpec:
        if folder_key is None and folder_path is not None:
            folder_key = self._folders_service.retrieve_key_by_folder_path(folder_path)
            folder_path = None

        return RequestSpec(
            method="POST",
            endpoint=Endpoint("/ecs_/v1/search"),
            content=json.dumps(
                {
                    "query": {"query": query, "numberOfResults": number_of_results},
                    "schema": {"name": name},
                }
            ),
            headers={
                **header_folder(folder_key, folder_path),
            },
        )



# type: ignore
import json
import os
import traceback
import uuid
from pathlib import Path
from typing import Optional

import click

from ._utils._input_args import generate_args
from ._utils._parse_ast import generate_bindings_json
from .middlewares import Middlewares


def generate_env_file(target_directory):
    env_path = os.path.join(target_directory, ".env")

    if not os.path.exists(env_path):
        relative_path = os.path.relpath(env_path, target_directory)
        click.echo(f"Created {relative_path} file.")
        with open(env_path, "w") as f:
            f.write("UIPATH_ACCESS_TOKEN=YOUR_TOKEN_HERE\n")
            f.write("UIPATH_URL=https://cloud.uipath.com/ACCOUNT_NAME/TENANT_NAME\n")


def get_user_script(directory: str, entrypoint: Optional[str] = None) -> Optional[str]:
    """Find the Python script to process."""
    if entrypoint:
        script_path = os.path.join(directory, entrypoint)
        if not os.path.isfile(script_path):
            click.echo(f"The {entrypoint} file does not exist in the current directory")
            return None
        return script_path

    python_files = [f for f in os.listdir(directory) if f.endswith(".py")]

    if not python_files:
        click.echo("No Python files found in the directory")
        return None
    elif len(python_files) == 1:
        return os.path.join(directory, python_files[0])
    else:
        click.echo(
            "Multiple Python files found in the current directory.\nPlease specify the entrypoint: `uipath init <entrypoint_path>`"
        )
        return None


@click.command()
@click.argument("entrypoint", required=False, default=None)
def init(entrypoint: str) -> None:
    """Initialize a uipath.json configuration file for the script."""
    current_directory = os.getcwd()
    generate_env_file(current_directory)

    result = Middlewares.next("init", entrypoint)

    if result.error_message:
        click.echo(result.error_message)
        if result.should_include_stacktrace:
            click.echo(traceback.format_exc())
        click.get_current_context().exit(1)

    if result.info_message:
        click.echo(result.info_message)

    if not result.should_continue:
        return

    script_path = get_user_script(current_directory, entrypoint=entrypoint)

    if not script_path:
        click.get_current_context().exit(1)

    try:
        args = generate_args(script_path)

        relative_path = Path(script_path).relative_to(current_directory).as_posix()

        config_data = {
            "entryPoints": [
                {
                    "filePath": relative_path,
                    "uniqueId": str(uuid.uuid4()),
                    # "type": "process", OR BE doesn't offer json schema support for type: Process
                    "type": "agent",
                    "input": args["input"],
                    "output": args["output"],
                }
            ]
        }

        # Generate bindings JSON based on the script path
        try:
            bindings_data = generate_bindings_json(script_path)

            # Add bindings to the config data
            config_data["bindings"] = bindings_data

            click.echo("Bindings generated successfully.")
        except Exception as e:
            click.echo(f"Warning: Could not generate bindings: {str(e)}")

        config_path = "uipath.json"
        with open(config_path, "w") as config_file:
            json.dump(config_data, config_file, indent=4)

        click.echo(f"Configuration file {config_path} created successfully.")

    except Exception as e:
        click.echo(f"Error generating configuration: {str(e)}")
        click.echo(traceback.format_exc())
        click.get_current_context().exit(1)



# type: ignore
import os

import click
import requests
from dotenv import load_dotenv


def get_most_recent_package():
    nupkg_files = [f for f in os.listdir(".uipath") if f.endswith(".nupkg")]
    if not nupkg_files:
        click.echo("No .nupkg file found in .uipath directory")
        return

    # Get full path and modification time for each file
    nupkg_files_with_time = [
        (f, os.path.getmtime(os.path.join(".uipath", f))) for f in nupkg_files
    ]

    # Sort by modification time (most recent first)
    nupkg_files_with_time.sort(key=lambda x: x[1], reverse=True)

    # Get most recent file
    return nupkg_files_with_time[0][0]


def get_env_vars():
    base_url = os.environ.get("UIPATH_URL")
    token = os.environ.get("UIPATH_ACCESS_TOKEN")

    if not all([base_url, token]):
        click.echo(
            "Missing required environment variables. Please check your .env file contains:"
        )
        click.echo("UIPATH_URL, UIPATH_ACCESS_TOKEN")
        raise click.Abort("Missing environment variables")

    return [base_url, token]


@click.command()
@click.option(
    "--tenant",
    "-t",
    "feed",
    flag_value="tenant",
    help="Whether to publish to the tenant package feed",
)
@click.option(
    "--personal-workspace",
    "-p",
    "feed",
    flag_value="personal",
    help="Whether to publish to the personal workspace",
)
def publish(feed):
    current_path = os.getcwd()
    load_dotenv(os.path.join(current_path, ".env"), override=True)
    if feed is None:
        click.echo("Select feed type:")
        click.echo("  0: Tenant package feed")
        click.echo("  1: Personal workspace")
        feed_idx = click.prompt("Select feed", type=int)
        feed = "tenant" if feed_idx == 0 else "personal"
        click.echo(f"Selected feed: {feed}")
    os.makedirs(".uipath", exist_ok=True)

    # Find most recent .nupkg file in .uipath directory
    most_recent = get_most_recent_package()

    if not most_recent:
        click.echo("Error: No package files found in .uipath directory")
        raise click.Abort()
    click.echo(f"Publishing most recent package: {most_recent}")

    package_to_publish_path = os.path.join(".uipath", most_recent)

    [base_url, token] = get_env_vars()

    url = f"{base_url}/orchestrator_/odata/Processes/UiPath.Server.Configuration.OData.UploadPackage()"

    if feed == "personal":
        # Get current user extended info to get personal workspace ID
        user_url = f"{base_url}/orchestrator_/odata/Users/UiPath.Server.Configuration.OData.GetCurrentUserExtended"
        user_response = requests.get(
            user_url, headers={"Authorization": f"Bearer {token}"}
        )

        if user_response.status_code != 200:
            click.echo("Failed to get user info")
            click.echo(f"Response: {user_response.text}")
            raise click.Abort()

        user_data = user_response.json()
        personal_workspace_id = user_data.get("PersonalWorskpaceFeedId")

        if not personal_workspace_id:
            click.echo("No personal workspace found for user")
            raise click.Abort()

        url = url + "?feedId=" + personal_workspace_id

    headers = {"Authorization": f"Bearer {token}"}

    with open(package_to_publish_path, "rb") as f:
        files = {"file": (package_to_publish_path, f, "application/octet-stream")}
        response = requests.post(url, headers=headers, files=files)

    if response.status_code == 200:
        click.echo("Package published successfully!")
    else:
        click.echo(f"Failed to publish package. Status code: {response.status_code}")
        click.echo(f"Response: {response.text}")



from ._endpoint import Endpoint
from ._infer_bindings import get_inferred_bindings_names, infer_bindings
from ._logs import setup_logging
from ._request_override import header_folder
from ._request_spec import RequestSpec
from ._user_agent import header_user_agent, user_agent_value

__all__ = [
    "Endpoint",
    "setup_logging",
    "RequestSpec",
    "header_folder",
    "get_inferred_bindings_names",
    "infer_bindings",
    "header_user_agent",
    "user_agent_value",
]



import logging
import sys
from typing import Optional

logger: logging.Logger = logging.getLogger("uipath")


def setup_logging(should_debug: Optional[bool] = None) -> None:
    logging.basicConfig(
        format="[%(asctime)s - %(name)s:%(lineno)d - %(levelname)s] %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    logger.setLevel(logging.DEBUG if should_debug else logging.INFO)
    logger.removeHandler(logging.StreamHandler(sys.stdout))
    logger.addHandler(logging.StreamHandler(sys.stderr))



import inspect
from logging import getLogger
from typing import Any, Union

from httpx import (
    URL,
    AsyncClient,
    Client,
    ConnectTimeout,
    Headers,
    Response,
    TimeoutException,
)
from tenacity import (
    retry,
    retry_if_exception,
    retry_if_result,
    wait_exponential,
)

from .._config import Config
from .._execution_context import ExecutionContext
from .._utils import user_agent_value
from .._utils.constants import HEADER_USER_AGENT


def is_retryable_exception(exception: BaseException) -> bool:
    return isinstance(exception, (ConnectTimeout, TimeoutException))


def is_retryable_status_code(response: Response) -> bool:
    return response.status_code >= 500 and response.status_code < 600


class BaseService:
    def __init__(self, config: Config, execution_context: ExecutionContext) -> None:
        self._logger = getLogger("uipath")
        self._config = config
        self._execution_context = execution_context
        self._tenant_scope_client = Client(
            base_url=self._config.base_url, headers=Headers(self.default_headers)
        )
        self._tenant_scope_client_async = AsyncClient(
            base_url=self._config.base_url, headers=Headers(self.default_headers)
        )
        org_scope_base_url = self.__get_org_scope_base_url()
        self._org_scope_client = Client(
            base_url=org_scope_base_url, headers=self.default_headers
        )
        self._org_scope_client_async = AsyncClient(
            base_url=org_scope_base_url, headers=self.default_headers
        )

        self._logger.debug(f"HEADERS: {self.default_headers}")

        super().__init__()

    @retry(
        retry=(
            retry_if_exception(is_retryable_exception)
            | retry_if_result(is_retryable_status_code)
        ),
        wait=wait_exponential(multiplier=1, min=1, max=10),
    )
    def request(
        self,
        method: str,
        url: Union[URL, str],
        **kwargs: Any,
    ) -> Response:
        self._logger.debug(f"Request: {method} {url}")
        self._logger.debug(
            f"HEADERS: {kwargs.get('headers', self._tenant_scope_client.headers)}"
        )

        try:
            stack = inspect.stack()

            # use the third frame because of the retry decorator
            caller_frame = stack[3].frame
            function_name = caller_frame.f_code.co_name

            if "self" in caller_frame.f_locals:
                module_name = type(caller_frame.f_locals["self"]).__name__
            elif "cls" in caller_frame.f_locals:
                module_name = caller_frame.f_locals["cls"].__name__
            else:
                module_name = ""
        except Exception:
            function_name = ""
            module_name = ""

        specific_component = (
            f"{module_name}.{function_name}" if module_name and function_name else ""
        )
        headers = kwargs.get("headers", {})
        headers[HEADER_USER_AGENT] = user_agent_value(specific_component)

        response = self._tenant_scope_client.request(method, url, **kwargs)
        response.raise_for_status()

        return response

    @retry(
        retry=(
            retry_if_exception(is_retryable_exception)
            | retry_if_result(is_retryable_status_code)
        ),
        wait=wait_exponential(multiplier=1, min=1, max=10),
    )
    async def request_async(
        self,
        method: str,
        url: Union[URL, str],
        **kwargs: Any,
    ) -> Response:
        self._logger.debug(f"Request: {method} {url}")
        self._logger.debug(
            f"HEADERS: {kwargs.get('headers', self._tenant_scope_client.headers)}"
        )

        try:
            stack = inspect.stack()

            # use the third frame because of the retry decorator
            caller_frame = stack[3].frame
            function_name = caller_frame.f_code.co_name

            if "self" in caller_frame.f_locals:
                module_name = type(caller_frame.f_locals["self"]).__name__
            elif "cls" in caller_frame.f_locals:
                module_name = caller_frame.f_locals["cls"].__name__
            else:
                module_name = ""
        except Exception:
            function_name = ""
            module_name = ""

        specific_component = (
            f"{module_name}.{function_name}" if module_name and function_name else ""
        )
        headers = kwargs.get("headers", {})
        headers[HEADER_USER_AGENT] = user_agent_value(specific_component)

        response = await self._tenant_scope_client_async.request(method, url, **kwargs)
        response.raise_for_status()

        return response

    def request_org_scope(
        self,
        method: str,
        url: Union[URL, str],
        **kwargs: Any,
    ) -> Response:
        self._logger.debug(f"Request: {method} {url}")
        self._logger.debug(
            f"HEADERS: {kwargs.get('headers', self._tenant_scope_client.headers)}"
        )

        try:
            stack = inspect.stack()

            # use the third frame because of the retry decorator
            caller_frame = stack[3].frame
            function_name = caller_frame.f_code.co_name

            if "self" in caller_frame.f_locals:
                module_name = type(caller_frame.f_locals["self"]).__name__
            elif "cls" in caller_frame.f_locals:
                module_name = caller_frame.f_locals["cls"].__name__
            else:
                module_name = ""
        except Exception:
            function_name = ""
            module_name = ""

        specific_component = (
            f"{module_name}.{function_name}" if module_name and function_name else ""
        )
        headers = kwargs.get("headers", {})
        headers[HEADER_USER_AGENT] = user_agent_value(specific_component)

        response = self._org_scope_client.request(method, url, **kwargs)
        response.raise_for_status()

        return response

    async def request_org_scope_async(
        self,
        method: str,
        url: Union[URL, str],
        **kwargs: Any,
    ) -> Response:
        self._logger.debug(f"Request: {method} {url}")
        self._logger.debug(
            f"HEADERS: {kwargs.get('headers', self._tenant_scope_client.headers)}"
        )

        try:
            stack = inspect.stack()

            # use the third frame because of the retry decorator
            caller_frame = stack[3].frame
            function_name = caller_frame.f_code.co_name

            if "self" in caller_frame.f_locals:
                module_name = type(caller_frame.f_locals["self"]).__name__
            elif "cls" in caller_frame.f_locals:
                module_name = caller_frame.f_locals["cls"].__name__
            else:
                module_name = ""
        except Exception:
            function_name = ""
            module_name = ""

        specific_component = (
            f"{module_name}.{function_name}" if module_name and function_name else ""
        )
        headers = kwargs.get("headers", {})
        headers[HEADER_USER_AGENT] = user_agent_value(specific_component)

        response = await self._org_scope_client_async.request(method, url, **kwargs)
        response.raise_for_status()

        return response

    @property
    def default_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            **self.auth_headers,
            **self.custom_headers,
        }

    @property
    def auth_headers(self) -> dict[str, str]:
        header = f"Bearer {self._config.secret}"
        return {"Authorization": header}

    @property
    def custom_headers(self) -> dict[str, str]:
        return {}

    def __get_org_scope_base_url(self) -> str:
        base_url = str(self._config.base_url)
        if base_url.endswith("/"):
            base_url = base_url[:-1]
        return base_url.rsplit("/", 1)[0] + "/"



import importlib

from .constants import HEADER_USER_AGENT


def user_agent_value(specific_component: str) -> str:
    product = "UiPath.Python.Sdk"
    product_component = f"UiPath.Python.Sdk.Activities.{specific_component}"

    version = importlib.metadata.version("uipath")

    return f"{product}/{product_component}/{version}"


def header_user_agent(specific_component: str) -> dict[str, str]:
    return {HEADER_USER_AGENT: user_agent_value(specific_component)}



import importlib
import inspect
import json
import logging
from functools import wraps
from typing import Any, Callable, List, Optional, Tuple

from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor

from ._otel_exporters import LlmOpsHttpExporter
from ._utils import _SpanUtils

logger = logging.getLogger(__name__)

trace.set_tracer_provider(TracerProvider())
trace.get_tracer_provider().add_span_processor(BatchSpanProcessor(LlmOpsHttpExporter()))  # type: ignore
tracer = trace.get_tracer(__name__)


class TracingManager:
    """Static utility class to manage tracing implementations and decorated functions."""

    # Registry to track original functions, decorated functions, and their parameters
    # Each entry is (original_func, decorated_func, params)
    _traced_registry: List[Tuple[Callable[..., Any], Callable[..., Any], Any]] = []

    # Custom tracer implementation
    _custom_tracer_implementation = None

    @classmethod
    def get_custom_tracer_implementation(cls):
        """Get the currently set custom tracer implementation."""
        return cls._custom_tracer_implementation

    @classmethod
    def register_traced_function(cls, original_func, decorated_func, params):
        """Register a function decorated with @traced and its parameters.

        Args:
            original_func: The original function before decoration
            decorated_func: The function after decoration
            params: The parameters used for tracing
        """
        cls._traced_registry.append((original_func, decorated_func, params))

    @classmethod
    def reapply_traced_decorator(cls, tracer_implementation):
        """Reapply a different tracer implementation to all functions previously decorated with @traced.

        Args:
            tracer_implementation: A function that takes the same parameters as _opentelemetry_traced
                                 and returns a decorator
        """
        cls._custom_tracer_implementation = tracer_implementation

        # Work with a copy of the registry to avoid modifying it during iteration
        registry_copy = cls._traced_registry.copy()

        for original_func, decorated_func, params in registry_copy:
            # Apply the new decorator with the same parameters
            new_decorated_func = tracer_implementation(**params)(original_func)

            logger.debug(
                f"Reapplying decorator to {original_func.__name__}, from {decorated_func.__name__}"
            )

            # If this is a method on a class, we need to update the class
            if hasattr(original_func, "__self__") and hasattr(
                original_func, "__func__"
            ):
                setattr(
                    original_func.__self__.__class__,
                    original_func.__name__,
                    new_decorated_func.__get__(
                        original_func.__self__, original_func.__self__.__class__
                    ),
                )
            else:
                # Replace the function in its module
                if hasattr(original_func, "__module__") and hasattr(
                    original_func, "__qualname__"
                ):
                    try:
                        module = importlib.import_module(original_func.__module__)
                        parts = original_func.__qualname__.split(".")

                        # Handle nested objects
                        obj = module
                        for part in parts[:-1]:
                            obj = getattr(obj, part)

                        setattr(obj, parts[-1], new_decorated_func)

                        # Update the registry entry for this function
                        # Find the index and replace with updated entry
                        for i, (orig, _dec, _p) in enumerate(cls._traced_registry):
                            if orig is original_func:
                                cls._traced_registry[i] = (
                                    original_func,
                                    new_decorated_func,
                                    params,
                                )
                                break
                    except (ImportError, AttributeError) as e:
                        # Log the error but continue processing other functions
                        logger.warning(f"Error reapplying decorator: {e}")
                        continue


def _default_input_processor(inputs):
    """Default input processor that doesn't log any actual input data."""
    return {"redacted": "Input data not logged for privacy/security"}


def _default_output_processor(outputs):
    """Default output processor that doesn't log any actual output data."""
    return {"redacted": "Output data not logged for privacy/security"}


def wait_for_tracers():
    """Wait for all tracers to finish."""
    trace.get_tracer_provider().shutdown()  # type: ignore


def _opentelemetry_traced(
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
):
    """Default tracer implementation using OpenTelemetry."""

    def decorator(func):
        @wraps(func)
        def sync_wrapper(*args, **kwargs):
            with tracer.start_as_current_span(func.__name__) as span:
                default_span_type = "function_call_sync"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                try:
                    result = func(*args, **kwargs)
                    # Process output if processor is provided
                    output = result
                    if output_processor is not None:
                        output = output_processor(result)
                    span.set_attribute("output", json.dumps(output, default=str))
                    return result
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        async def async_wrapper(*args, **kwargs):
            with tracer.start_as_current_span(func.__name__) as span:
                default_span_type = "function_call_async"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                try:
                    result = await func(*args, **kwargs)
                    # Process output if processor is provided
                    output = result
                    if output_processor is not None:
                        output = output_processor(result)
                    span.set_attribute("output", json.dumps(output, default=str))
                    return result
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        def generator_wrapper(*args, **kwargs):
            with tracer.start_as_current_span(func.__name__) as span:
                default_span_type = "function_call_generator_sync"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                outputs = []
                try:
                    for item in func(*args, **kwargs):
                        outputs.append(item)
                        span.add_event(f"Yielded: {item}")  # Add event for each yield
                        yield item

                    # Process output if processor is provided
                    output_to_record = outputs
                    if output_processor is not None:
                        output_to_record = output_processor(outputs)
                    span.set_attribute(
                        "output", json.dumps(output_to_record, default=str)
                    )
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        @wraps(func)
        async def async_generator_wrapper(*args, **kwargs):
            with tracer.start_as_current_span(func.__name__) as span:
                default_span_type = "function_call_generator_async"
                span.set_attribute(
                    "span_type",
                    span_type if span_type is not None else default_span_type,
                )
                if run_type is not None:
                    span.set_attribute("run_type", run_type)

                # Format arguments for tracing
                inputs = _SpanUtils.format_args_for_trace_json(
                    inspect.signature(func), *args, **kwargs
                )
                # Apply input processor if provided
                if input_processor is not None:
                    processed_inputs = input_processor(json.loads(inputs))
                    inputs = json.dumps(processed_inputs, default=str)
                span.set_attribute("inputs", inputs)
                outputs = []
                try:
                    async for item in func(*args, **kwargs):
                        outputs.append(item)
                        span.add_event(f"Yielded: {item}")  # Add event for each yield
                        yield item

                    # Process output if processor is provided
                    output_to_record = outputs
                    if output_processor is not None:
                        output_to_record = output_processor(outputs)
                    span.set_attribute(
                        "output", json.dumps(output_to_record, default=str)
                    )
                except Exception as e:
                    span.record_exception(e)
                    span.set_status(
                        trace.status.Status(trace.status.StatusCode.ERROR, str(e))
                    )
                    raise

        if inspect.iscoroutinefunction(func):
            return async_wrapper
        elif inspect.isgeneratorfunction(func):
            return generator_wrapper
        elif inspect.isasyncgenfunction(func):
            return async_generator_wrapper
        else:
            return sync_wrapper

    return decorator


def traced(
    run_type: Optional[str] = None,
    span_type: Optional[str] = None,
    input_processor: Optional[Callable[..., Any]] = None,
    output_processor: Optional[Callable[..., Any]] = None,
    hide_input: bool = False,
    hide_output: bool = False,
):
    """Decorator that will trace function invocations.

    Args:
        run_type: Optional string to categorize the run type
        span_type: Optional string to categorize the span type
        input_processor: Optional function to process function inputs before recording
            Should accept a dictionary of inputs and return a processed dictionary
        output_processor: Optional function to process function outputs before recording
            Should accept the function output and return a processed value
        hide_input: If True, don't log any input data
        hide_output: If True, don't log any output data
    """
    # Apply default processors selectively based on hide flags
    if hide_input:
        input_processor = _default_input_processor
    if hide_output:
        output_processor = _default_output_processor

    # Store the parameters for later reapplication
    params = {
        "run_type": run_type,
        "span_type": span_type,
        "input_processor": input_processor,
        "output_processor": output_processor,
    }

    # Check for custom implementation first
    custom_implementation = TracingManager.get_custom_tracer_implementation()
    tracer_impl: Any = (
        custom_implementation if custom_implementation else _opentelemetry_traced
    )

    def decorator(func):
        # Decorate the function
        decorated_func = tracer_impl(**params)(func)
        # Register both original and decorated function with parameters
        TracingManager.register_traced_function(func, decorated_func, params)
        return decorated_func

    return decorator



import os
import time
from typing import Optional

import click
import requests

from ._models import TenantsAndOrganizationInfoResponse, TokenData
from ._oidc_utils import get_auth_config
from ._utils import (
    get_auth_data,
    get_parsed_token_data,
    update_auth_file,
    update_env_file,
)


class PortalService:
    """Service for interacting with the UiPath Portal API."""

    access_token: Optional[str] = None
    prt_id: Optional[str] = None
    domain: Optional[str] = None
    selected_tenant: Optional[str] = None

    _tenants_and_organizations: Optional[TenantsAndOrganizationInfoResponse] = None

    def __init__(
        self,
        domain: str,
        access_token: Optional[str] = None,
        prt_id: Optional[str] = None,
    ):
        self.domain = domain
        self.access_token = access_token
        self.prt_id = prt_id

    def update_token_data(self, token_data: TokenData):
        self.access_token = token_data["access_token"]
        self.prt_id = get_parsed_token_data(token_data).get("prt_id")

    def get_tenants_and_organizations(self) -> TenantsAndOrganizationInfoResponse:
        url = f"https://{self.domain}.uipath.com/{self.prt_id}/portal_/api/filtering/leftnav/tenantsAndOrganizationInfo"
        response = requests.get(
            url, headers={"Authorization": f"Bearer {self.access_token}"}
        )
        if response.ok:
            result = response.json()
            self._tenants_and_organizations = result
            return result
        elif response.status_code == 401:
            raise Exception("Unauthorized")
        else:
            raise Exception(
                f"Failed to get tenants and organizations: {response.status_code} {response.text}"
            )

    def get_uipath_orchestrator_url(self) -> str:
        if self._tenants_and_organizations is None:
            self._tenants_and_organizations = self.get_tenants_and_organizations()
        organization = self._tenants_and_organizations.get("organization")
        if organization is None:
            raise Exception("Organization not found")
        account_name = organization.get("name")
        return f"https://{self.domain}.uipath.com/{account_name}/{self.selected_tenant}/orchestrator_"

    def post_refresh_token_request(self, refresh_token: str) -> TokenData:
        url = f"https://{self.domain}.uipath.com/identity_/connect/token"
        client_id = get_auth_config().get("client_id")

        data = {
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
            "client_id": client_id,
        }

        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        response = requests.post(url, data=data, headers=headers)
        if response.ok:
            return response.json()
        elif response.status_code == 401:
            raise Exception("Unauthorized")
        else:
            raise Exception(f"Failed to refresh token: {response.status_code}")

    def ensure_valid_token(self):
        """Ensure the access token is valid and refresh it if necessary.

        This function should be called when running CLI commands to verify authentication.
        It checks if an auth file exists and contains a valid non-expired token.
        If the token is expired, it will attempt to refresh it.
        If no auth file exists, it will raise an exception.

        Raises:
            Exception: If no auth file exists or token refresh fails
        """
        auth_data = get_auth_data()
        claims = get_parsed_token_data(auth_data)
        exp = claims.get("exp")

        if exp is not None and float(exp) > time.time():
            if not os.getenv("UIPATH_URL"):
                tenants_and_organizations = self.get_tenants_and_organizations()
                select_tenant(
                    self.domain if self.domain else "alpha", tenants_and_organizations
                )
            return auth_data.get("access_token")

        refresh_token = auth_data.get("refresh_token")
        if refresh_token is None:
            raise Exception("Refresh token not found")
        token_data = self.post_refresh_token_request(refresh_token)
        update_auth_file(token_data)

        self.access_token = token_data["access_token"]
        self.prt_id = claims.get("prt_id")

        updated_env_contents = {
            "UIPATH_ACCESS_TOKEN": token_data["access_token"],
        }
        if not os.getenv("UIPATH_URL"):
            tenants_and_organizations = self.get_tenants_and_organizations()
            select_tenant(
                self.domain if self.domain else "alpha", tenants_and_organizations
            )

        update_env_file(updated_env_contents)

    def has_initialized_auth(self):
        try:
            auth_data = get_auth_data()
            if not auth_data or "access_token" not in auth_data:
                return False
            if not os.path.exists(".env"):
                return False
            if not os.getenv("UIPATH_ACCESS_TOKEN"):
                return False

            return True
        except Exception:
            return False


def select_tenant(
    domain: str, tenants_and_organizations: TenantsAndOrganizationInfoResponse
):
    tenant_names = [tenant["name"] for tenant in tenants_and_organizations["tenants"]]
    click.echo("Available tenants:")
    for idx, name in enumerate(tenant_names):
        click.echo(f"  {idx}: {name}")
    tenant_idx = (
        0 if len(tenant_names) == 1 else click.prompt("Select tenant", type=int)
    )
    tenant_name = tenant_names[tenant_idx]
    account_name = tenants_and_organizations["organization"]["name"]
    click.echo(f"Selected tenant: {tenant_name}")

    update_env_file(
        {
            "UIPATH_URL": f"https://{domain if domain else 'alpha'}.uipath.com/{account_name}/{tenant_name}",
            "UIPATH_TENANT_ID": tenants_and_organizations["tenants"][tenant_idx]["id"],
            "UIPATH_ORGANIZATION_ID": tenants_and_organizations["organization"]["id"],
        }
    )



import click


def environment_options(function):
    function = click.option(
        "--alpha",
        "domain",
        flag_value="alpha",
        help="Use alpha environment",
    )(function)
    function = click.option(
        "--staging",
        "domain",
        flag_value="staging",
        help="Use staging environment",
    )(function)
    function = click.option(
        "--cloud",
        "domain",
        flag_value="cloud",
        default=True,
        help="Use production environment",
    )(function)
    return function



# type: ignore
import json
import os
import socket
import webbrowser

import click
from dotenv import load_dotenv

from ._auth._auth_server import HTTPSServer
from ._auth._oidc_utils import get_auth_config, get_auth_url
from ._auth._portal_service import PortalService, select_tenant
from ._auth._utils import update_auth_file, update_env_file
from ._utils._common import environment_options

load_dotenv()


def is_port_in_use(port: int) -> bool:
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("localhost", port))
            s.close()
            return False
        except socket.error:
            return True


def set_port():
    auth_config = get_auth_config()
    port = auth_config.get("port", 8104)
    port_option_one = auth_config.get("portOptionOne", 8104)
    port_option_two = auth_config.get("portOptionTwo", 8055)
    port_option_three = auth_config.get("portOptionThree", 42042)
    if is_port_in_use(port):
        if is_port_in_use(port_option_one):
            if is_port_in_use(port_option_two):
                if is_port_in_use(port_option_three):
                    raise RuntimeError(
                        "All configured ports are in use. Please close applications using ports or configure different ports."
                    )
                else:
                    port = port_option_three
            else:
                port = port_option_two
        else:
            port = port_option_one
    auth_config["port"] = port
    with open(
        os.path.join(os.path.dirname(__file__), "..", "auth_config.json"), "w"
    ) as f:
        json.dump(auth_config, f)


@click.command()
@environment_options
def auth(domain="alpha"):
    """Authenticate with UiPath Cloud Platform."""
    portal_service = PortalService(domain)
    if (
        os.getenv("UIPATH_URL")
        and os.getenv("UIPATH_TENANT_ID")
        and os.getenv("UIPATH_ORGANIZATION_ID")
    ):
        try:
            portal_service.ensure_valid_token()
            click.echo("Authentication successful")
            return
        except Exception:
            click.echo(
                "Authentication not found or expired. Please authenticate again."
            )

    auth_url, code_verifier, state = get_auth_url(domain)

    webbrowser.open(auth_url, 1)
    auth_config = get_auth_config()

    print(
        "If a browser window did not open, please open the following URL in your browser:"
    )
    print(auth_url)
    server = HTTPSServer(port=auth_config["port"])
    token_data = server.start(state, code_verifier, domain)
    try:
        if token_data:
            portal_service.update_token_data(token_data)
            update_auth_file(token_data)
            access_token = token_data["access_token"]
            update_env_file({"UIPATH_ACCESS_TOKEN": access_token})

            tenants_and_organizations = portal_service.get_tenants_and_organizations()
            select_tenant(domain, tenants_and_organizations)
        else:
            click.echo("Authentication failed")
    except Exception as e:
        click.echo(f"Authentication failed: {e}")



from functools import wraps

from uipath.tracing._traced import TracingManager, traced


# Custom wrapper that does nothing
def donothing_custom_tracer(**kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Simple implementation that just adds a marker to the result
            result = func(*args, **kwargs)
            return result

        return wrapper

    return decorator


# Helper function for testing custom tracer
def simple_custom_tracer(**kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Simple implementation that just adds a marker to the result
            result = func(*args, **kwargs)
            if isinstance(result, dict):
                result["custom_tracer_used"] = True
            return result

        return wrapper

    return decorator


# Helper function for testing custom tracer with method
def custom_method_tracer(**kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = func(*args, **kwargs)
            if isinstance(result, dict):
                result["custom_method_tracer_used"] = True
            return result

        return wrapper

    return decorator


# Helper function for testing with counter
def custom_tracer_with_counter(call_counter, **kwargs):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            call_counter["count"] += 1
            result = func(*args, **kwargs)
            if isinstance(result, dict):
                result["custom_tracer_id"] = call_counter["count"]
            return result

        return wrapper

    return decorator


# Define the test classes
class TestClassForMethodTest:
    @traced()
    def sample_method(self, x, y):
        return {"product": x * y}


# Module level functions for function test
@traced()
def func1_for_test(x):
    return {"result": x * 2}


@traced()
def func2_for_test(x):
    return {"result": x * 3}


# Define a function with @traced
@traced()
def sample_function():
    return {"status": "success"}


def test_tracing_manager_custom_implementation():
    """Test setting and getting a custom tracer implementation."""
    # Set the custom implementation
    TracingManager.reapply_traced_decorator(simple_custom_tracer)

    # Get the implementation and verify it's the same one
    impl = TracingManager.get_custom_tracer_implementation()
    assert impl is simple_custom_tracer


def test_traced_with_custom_implementation():
    """Test that @traced uses a custom implementation when provided."""
    # Set the custom implementation
    TracingManager.reapply_traced_decorator(simple_custom_tracer)

    # Call the function and verify the custom implementation was used
    result = sample_function()
    assert "custom_tracer_used" in result
    assert result["custom_tracer_used"] is True


def test_reapply_traced_decorator_to_class_method():
    """Test reapply_traced_decorator with class methods."""
    TracingManager.reapply_traced_decorator(donothing_custom_tracer)

    # Create instance and call with default implementation
    instance = TestClassForMethodTest()
    result1 = instance.sample_method(2, 3)
    assert result1 == {"product": 6}

    # Apply our custom implementation
    TracingManager.reapply_traced_decorator(custom_method_tracer)

    # Create a NEW instance which will use the updated class method
    new_instance = TestClassForMethodTest()
    result2 = new_instance.sample_method(2, 3)

    # Verify the result
    assert "product" in result2
    assert result2["product"] == 6
    assert "custom_method_tracer_used" in result2


def test_reapply_to_module_level_functions():
    """Test reapply_traced_decorator with module level functions."""

    TracingManager.reapply_traced_decorator(donothing_custom_tracer)

    # First call with default implementation
    assert func1_for_test(5) == {"result": 10}
    assert func2_for_test(5) == {"result": 15}

    # Counter to track custom tracer calls
    call_counter = {"count": 0}

    # Reapply with the custom implementation
    TracingManager.reapply_traced_decorator(
        lambda **kwargs: custom_tracer_with_counter(call_counter, **kwargs)
    )

    # Call the functions directly - they should now use the updated implementation
    result1 = func1_for_test(5)
    result2 = func2_for_test(5)

    # Verify the custom implementation was applied
    assert "result" in result1
    assert result1["result"] == 10
    assert "custom_tracer_id" in result1

    assert "result" in result2
    assert result2["result"] == 15
    assert "custom_tracer_id" in result2

    # Verify both functions were processed
    assert call_counter["count"] == 2



import pydantic_core
import pytest

from uipath import UiPath


class TestSdkConfig:
    def test_no_config(self, monkeypatch):
        monkeypatch.delenv("UIPATH_URL", raising=False)
        monkeypatch.delenv("UIPATH_ACCESS_TOKEN", raising=False)
        monkeypatch.delenv("UNATTENDED_USER_ACCESS_TOKEN", raising=False)

        with pytest.raises(pydantic_core._pydantic_core.ValidationError) as exc_info:
            UiPath()

        assert exc_info.value.errors(include_url=False) == [
            {
                "type": "string_type",
                "loc": ("base_url",),
                "msg": "Input should be a valid string",
                "input": None,
            },
            {
                "type": "string_type",
                "loc": ("secret",),
                "msg": "Input should be a valid string",
                "input": None,
            },
        ]

    def test_config_from_env(self, monkeypatch):
        monkeypatch.setenv("UIPATH_URL", "https://example.com")
        monkeypatch.setenv("UIPATH_ACCESS_TOKEN", "1234567890")
        sdk = UiPath()
        assert sdk._config.base_url == "https://example.com"
        assert sdk._config.secret == "1234567890"

    def test_config_from_constructor(self, monkeypatch):
        monkeypatch.delenv("UIPATH_URL", raising=False)
        monkeypatch.delenv("UIPATH_ACCESS_TOKEN", raising=False)
        monkeypatch.delenv("UNATTENDED_USER_ACCESS_TOKEN", raising=False)

        sdk = UiPath(base_url="https://example.com", secret="1234567890")
        assert sdk._config.base_url == "https://example.com"
        assert sdk._config.secret == "1234567890"



from pydantic import BaseModel


class ContextGroundingMetadata(BaseModel):
    operation_id: str
    strategy: str


class ContextGroundingQueryResponse(BaseModel):
    id: str
    reference: str
    source: str
    page_number: str
    source_document_id: str
    caption: str
    score: float
    content: str
    metadata: ContextGroundingMetadata
